package org.apache.commons.text.link;

import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.PrintStream;

/**
 * TODO add javadoc
 *
 */
public class Parser {
    /* This is a hack that allows one to discard disjuncts containing
       connectors whose cost is greater than given a bound. This number plus
       the cost of any connectors on a disjunct must remain negative, and
       this number multiplied times the number of costly connectors on any
       disjunct must fit into an integer. */
    final static int NEGATIVECOST = -1000000;

    final static int NOCUTOFF = 1000; /* no connector will have cost this high */

    /* the string to use to show the wall */
    final static String LEFT_WALL_DISPLAY = "LEFT-WALL";

    /* If this connector is used on the wall, */
    /* then suppress the display of the wall  */
    /* bogus name to prevent ever suppressing */
    final static String LEFT_WALL_SUPPRESS = "Wd";

    /* the string to use to show the wall */
    final static String RIGHT_WALL_DISPLAY = "RIGHT-WALL";

    /* If this connector is used on the wall, */
    final static String RIGHT_WALL_SUPPRESS = "RW";

    /* The following define the names of the special strings in the dictionary. */
    final static String LEFT_WALL_WORD = "LEFT-WALL";
    final static String RIGHT_WALL_WORD = "RIGHT-WALL";
    final static String POSTPROCESS_WORD = "POSTPROCESS";
    final static String ANDABLE_CONNECTORS_WORD = "ANDABLE-CONNECTORS";
    final static String UNLIMITED_CONNECTORS_WORD = "UNLIMITED-CONNECTORS";

    final static String PROPER_WORD = "CAPITALIZED-WORDS";
    final static String PL_PROPER_WORD = "PL-CAPITALIZED-WORDS";
    final static String HYPHENATED_WORD = "HYPHENATED-WORDS";
    final static String NUMBER_WORD = "NUMBERS";
    final static String ING_WORD = "ING-WORDS";
    final static String S_WORD = "S-WORDS";
    final static String ED_WORD = "ED-WORDS";
    final static String LY_WORD = "LY-WORDS";
    final static String UNKNOWN_WORD = "UNKNOWN-WORD";

    /* file names (including paths)
     should not be longer than this */
    final static int MAX_PATH_NAME = 200;

    /*      Some size definitions.  Reduce these for small machines */
    final static int MAX_WORD = 60; /* maximum number of chars in a word */
    final static int MAX_LINE = 1500; /* maximum number of chars in a sentence */
    final static int MAX_SENTENCE = 250; /* maximum number of words in a sentence */

    /* This cannot be more than 254, because I use word MAX_SENTENCE+1 to
       indicate that nothing can connect to this connector, and this
       should fit in one byte (if the word field of a connector is an
       unsigned char) */
    final static int MAX_LINKS = (2 * MAX_SENTENCE - 3); /* maximum number of links allowed */
    final static int MAX_TOKEN_LENGTH = 50; /* maximum number of chars in a token */
    final static int MAX_DISJUNCT_COST = 10000;

    final static int DOWN_priority = 2;
    final static int UP_priority = 1;
    final static int THIN_priority = 0;

    /* the labels >= 0 are used by fat links */
    final static int NORMAL_LABEL = (-1); /* used for normal connectors            */

    final static int UNLIMITED_LEN = 255;
    final static int SHORT_LEN = 6;
    final static int NO_WORD = 255;

    /* Here are the types */
    final static int OR_type = 0;
    final static int AND_type = 1;
    final static int CONNECTOR_type = 2;

    /* These parameters tell power_pruning, to tell whether this is before or after
       generating and disjuncts.  GENTLE is before RUTHLESS is after. */
    final static int GENTLE = 1;
    final static int RUTHLESS = 0;

    static final int PP_LEXER_MAX_LABELS = 512;

    static final int VDAL = 1; /* CostModel sort by Violations, Disjunct cost, And cost, Link cost */

    final static int HT_SIZE = (1 << 10);

    /* size of random table for computing the
       hash functions.  must be a power of 2 */
    final static int RTSIZE = 256;

    final static int NODICT = 1;
    final static int DICTPARSE = 2;
    final static int WORDFILE = 3;
    final static int SEPARATE = 4;
    final static int NOTINDICT = 5;
    final static int BUILDEXPR = 6;
    final static int INTERNALERROR = 7;

    final static int LINKSET_SPARSENESS = 2;
    final static int LINKSET_MAX_SETS = 512;
    final static int LINKSET_DEFAULT_SEED = 37;

    final static int PP_FIRST_PASS = 1;
    final static int PP_SECOND_PASS = 2;

    final static int MAXINPUT = 1024;
    final static int DISPLAY_MAX = 1024;
    final static char COMMENT_CHAR = '%'; /* input lines beginning with this are ignored */

    final static char UNGRAMMATICAL = '*';
    final static char PARSE_WITH_DISJUNCT_COST_GT_0 = ':';
    final static char NO_LABEL = ' ';

    /* the indiction in a word field that this connector cannot
       be used -- is obsolete.
    */
    final static int BAD_WORD = (MAX_SENTENCE + 1);

    final static int PP_MAX_DOMAINS = 128;

    //final static int LINKSET_SPARSENESS=2;
    final static int LINKSET_SEED_VALUE = 37;

    final static int PP_MAX_UNIQUE_LINK_NAMES = 1024; /* just needs to be approximate */

    final static int LINE_LIMIT = 70;

    static final String DEFAULTPATH = ".:./data:./data/link:";

    final static int MAX_STRIP = 10;

    final static int MAX_HEIGHT = 30;

    final static int COMMA_LABEL = (-2); /* to hook the comma to the following "and" */
    final static int EITHER_LABEL = (-3); /* to connect the "either" to the following "or" */
    final static int NEITHER_LABEL = (-4); /* to connect the "neither" to the following "nor"*/
    final static int NOT_LABEL = (-5); /* to connect the "not" to the following "but"*/
    final static int NOTONLY_LABEL = (-6); /* to connect the "not" to the following "only"*/
    final static int BOTH_LABEL = (-7); /* to connect the "both" to the following "and"*/

    final static int MAXCONSTITUENTS = 1024;
    final static int MAXSUBL = 16;
    final static char OPEN_BRACKET = '[';
    final static char CLOSE_BRACKET = ']';

    final static int CType_OPEN = 0;
    final static int CType_CLOSE = 1;
    final static int CType_WORD = 2;

    final static int WType_NONE = 0;
    final static int WType_STYPE = 1;
    final static int WType_PTYPE = 2;
    final static int WType_QTYPE = 3;
    final static int WType_QDTYPE = 4;

    static int batch_errors = 0;
    static boolean input_pending = false;
    static int input_char;
    static ParseOptions opts;

    static int lperrno;
    static String lperrmsg;

    static int STAT_N_disjuncts; /* keeping statistics */
    static int STAT_calls_to_equality_test;

    public static void main(String arg[]) throws IOException {

        Dictionary dict;
        Sentence sent;
        String dictionary_file = null;
        String post_process_knowledge_file = null;
        String constituent_knowledge_file = null;
        String affix_file = null;
        boolean pp_on = true;
        boolean af_on = true;
        boolean cons_on = true;
        int num_linkages, i;
        StringBuffer input_string = new StringBuffer();
        int label = NO_LABEL;

        i = 0;
        if (arg.length > 1 && (arg[0].charAt(0) != '-')) {
            /* the dictionary is the first argument if it doesn't begin with "-" */
            dictionary_file = arg[0];
            i++;
        }

        opts = new ParseOptions();

        opts.parse_options_set_max_sentence_length(70);
        opts.parse_options_set_linkage_limit(1000);
        opts.parse_options_set_short_length(10);

        for (; i < arg.length; i++) {
            if (arg[i].charAt(0) == '-') {
                if (arg[i].equals("-pp")) {
                    if ((post_process_knowledge_file != null) || (i + 1 == arg.length))
                        print_usage(arg[0]);
                    post_process_knowledge_file = arg[i + 1];
                    i++;
                } else if (arg[i].equals("-c")) {
                    if ((constituent_knowledge_file != null) || (i + 1 == arg.length))
                        print_usage(arg[0]);
                    constituent_knowledge_file = arg[i + 1];
                    i++;
                } else if (arg[i].equals("-a")) {
                    if ((affix_file != null) || (i + 1 == arg.length))
                        print_usage(arg[0]);
                    affix_file = arg[i + 1];
                    i++;
                } else if (arg[i].equals("-ppoff")) {
                    pp_on = false;
                } else if (arg[i].equals("-coff")) {
                    cons_on = false;
                } else if (arg[i].equals("-aoff")) {
                    af_on = false;
                } else if (arg[i].equals("-batch")) {
                    if ((opts.input != System.in) || (i + 1 == arg.length))
                        print_usage(arg[0]);
                    opts.input = new FileInputStream(arg[i + 1]);
                    i++;
                } else if (arg[i].equals("-out")) {
                    if ((opts.out != System.out) || (i + 1 == arg.length))
                        print_usage(arg[0]);
                    opts.out = new PrintStream(new FileOutputStream(arg[i + 1]));
                    i++;
                } else if (arg[i].charAt(1) == '!') {
                } else {
                    print_usage(arg[0]);
                }
            } else {
                print_usage(arg[0]);
            }
        }

        if (!pp_on && post_process_knowledge_file != null)
            print_usage(arg[0]);

        if (dictionary_file == null) {
            dictionary_file = "4.0.dict";
            // TODO: logging
            System.err.println("No dictionary file specified.  Using " + dictionary_file + ".");
        }

        if (af_on && affix_file == null) {
            affix_file = "4.0.affix";
            // TODO: logging
            System.err.println("No affix file specified.  Using " + affix_file + ".");
        }

        if (pp_on && post_process_knowledge_file == null) {
            post_process_knowledge_file = "4.0.knowledge";
            // TODO: logging
            System.err.println("No post process knowledge file specified.  Using " + post_process_knowledge_file + ".");
        }

        if (cons_on && constituent_knowledge_file == null) {
            constituent_knowledge_file = "4.0.constituent-knowledge";
            // TODO: logging
            System.err.println("No constituent knowledge file specified.  Using " + constituent_knowledge_file + ".");
        }

        dict =
            new Dictionary(opts, dictionary_file, post_process_knowledge_file, constituent_knowledge_file, affix_file);

        /* process the command line like commands */
        for (i = 1; i < arg.length; i++) {
            if (!arg[i].equals("-pp") && !arg[i].equals("-c") && !arg[i].equals("-a")) {
                i++;
            } else if (
                arg[i].charAt(0) == '-'
                    && !arg[i].equals("-ppoff")
                    && !arg[i].equals("-coff")
                    && !arg[i].equals("-aoff")) {
                opts.issue_special_command(arg[i].substring(1), dict);
            }
        }

        while (fget_input_string(input_string, opts.input, opts.out, opts)) {
            if (input_string.length() == 0) {
                continue;
            }
            if (input_string.equals("quit\n") || input_string.equals("exit\n"))
                break;
            if (special_command(input_string, dict))
                continue;
            if (opts.parse_options_get_echo_on()) {
                opts.out.println(input_string);
            }

            if (opts.parse_options_get_batch_mode()) {
                label = strip_off_label(input_string);
            }

            sent = new Sentence(input_string.toString(), dict, opts);
            if (sent.sentence_length() > opts.parse_options_get_max_sentence_length()) {
                if (opts.verbosity > 0) {
                    opts.out.println(
                        "Sentence length ("
                            + sent.sentence_length()
                            + " words) exceeds maximum allowable ("
                            + opts.parse_options_get_max_sentence_length()
                            + " words)");
                }
                continue;
            } /* First parse with cost 0 or 1 and no null links */
            opts.parse_options_set_disjunct_cost(2);
            opts.parse_options_set_min_null_count(0);
            opts.parse_options_set_max_null_count(0);
            opts.parse_options_reset_resources();
            num_linkages = sent.sentence_parse(opts);
            /* Now parse with null links */
            if (num_linkages == 0 && !opts.parse_options_get_batch_mode()) {
                if (opts.verbosity > 0)
                    opts.out.println("No complete linkages found.");
                if (opts.parse_options_get_allow_null()) {
                    opts.parse_options_set_min_null_count(1);
                    opts.parse_options_set_max_null_count(sent.sentence_length());
                    num_linkages = sent.sentence_parse(opts);
                }
            }

            opts.print_total_time();
            if (opts.parse_options_get_batch_mode()) {
                batch_process_some_linkages(label, sent, opts);
            } else {
                process_some_linkages(sent, opts);
            }

        }

        if (opts.parse_options_get_batch_mode()) {
            opts.print_time("Total");
            opts.out.println("" + batch_errors + " error" + ((batch_errors == 1) ? "" : "s") + ".");
        }
    }

    static void print_usage(String str) {
        System.err.println(
            "Usage: <class name> [dict_file] [-pp PPKnowledge_file]\n"
                + "          [-c constituent_knowledge_file] [-a affix_file]\n"
                + "          [-ppoff] [-coff] [-aoff] [-batch] [-<special \"!\" command>]");
        System.exit(1);
    }

    static void left_append_string(StringBuffer string, String s, String t) {
        /* prints s then prints the last |t|-|s| characters of t.
           if s is longer than t, it truncates s.
        */
        int i, j, k;
        j = t.length();
        k = s.length();
        for (i = 0; i < j; i++) {
            if (i < k) {
                string.append(s.charAt(i));
            } else {
                string.append(t.charAt(i));
            }
        }
    }

    static int strip_off_label(StringBuffer input_string) {
        int c;

        c = input_string.charAt(0);
        switch (c) {
            case UNGRAMMATICAL :
            case PARSE_WITH_DISJUNCT_COST_GT_0 :
                input_string.setCharAt(0, ' ');
                return c;
            default :
                return NO_LABEL;
        }
    }

    static boolean special_command(StringBuffer input_string, Dictionary dict) {

        if (input_string.charAt(0) == '\n')
            return true;
        if (input_string.charAt(0) == COMMENT_CHAR)
            return true;
        if (input_string.charAt(0) == '!') {
            opts.issue_special_command(input_string.substring(1), dict);
            return true;
        }
        return false;
    }

    static void batch_process_some_linkages(int label, Sentence sent, ParseOptions opts) {
        Linkage linkage;

        if (there_was_an_error(label, sent, opts) != 0) {
            if (sent.sentence_num_linkages_found() > 0) {
                linkage = new Linkage(0, sent, opts);
                linkage.process_linkage(opts);
            }
            opts.out.println("+++++ error " + batch_errors);
        }
    }

    static int there_was_an_error(int label, Sentence sent, ParseOptions opts) {

        if (sent.sentence_num_valid_linkages() > 0) {
            if (label == UNGRAMMATICAL) {
                opts.out.println("error: parsed ungrammatical sentence");
                batch_errors++;
                return UNGRAMMATICAL;
            }
            if ((sent.sentence_disjunct_cost(0) == 0) && (label == PARSE_WITH_DISJUNCT_COST_GT_0)) {
                opts.out.println("error: cost=0");
                batch_errors++;
                return PARSE_WITH_DISJUNCT_COST_GT_0;
            }
        } else {
            if (label != UNGRAMMATICAL) {
                opts.out.println("error: failed");
                batch_errors++;
                return UNGRAMMATICAL;
            }
        }
        return 0;
    }

    static void process_some_linkages(Sentence sent, ParseOptions opts) throws IOException {
        int i, c, num_displayed, num_to_query;
        Linkage linkage;

        if (opts.verbosity > 0)
            sent.print_parse_statistics(opts);
        if (!opts.parse_options_get_display_bad()) {
            num_to_query = Math.min(sent.sentence_num_valid_linkages(), DISPLAY_MAX);
        } else {
            num_to_query = Math.min(sent.sentence_num_linkages_post_processed(), DISPLAY_MAX);
        }

        for (i = 0, num_displayed = 0; i < num_to_query; ++i) {

            if ((sent.sentence_num_violations(i) > 0) && (!opts.parse_options_get_display_bad())) {
                continue;
            }

            linkage = new Linkage(i, sent, opts);

            if (opts.verbosity > 0) {
                if (sent.sentence_num_valid_linkages() == 1 && (!opts.parse_options_get_display_bad())) {
                    opts.out.print("  Unique linkage, ");
                } else if ((opts.parse_options_get_display_bad()) && (sent.sentence_num_violations(i) > 0)) {
                    opts.out.print("  Linkage " + (i + 1) + " (bad), ");
                } else {
                    opts.out.print("  Linkage " + (i + 1) + ", ");
                }

                if (!linkage.linkage_is_canonical()) {
                    opts.out.print("non-canonical, ");
                }
                if (linkage.linkage_is_improper()) {
                    opts.out.print("improper fat linkage, ");
                }
                if (linkage.linkage_has_inconsistent_domains()) {
                    opts.out.print("inconsistent domains, ");
                }
                opts.out.println(
                    "cost vector = (UNUSED="
                        + linkage.linkage_unused_word_cost()
                        + " DIS="
                        + linkage.linkage_disjunct_cost()
                        + " AND="
                        + linkage.linkage_and_cost()
                        + " LEN="
                        + linkage.linkage_link_cost()
                        + ")");
            }

            linkage.process_linkage(opts);

            if (++num_displayed < num_to_query) {
                if (opts.verbosity > 0) {
                    opts.out.println("Press RETURN for the next linkage.");
                }
                c = fget_input_char(System.in, opts);
                if (c != '\n' && c != '\r') {
                    input_char = c;
                    input_pending = true;
                    break;
                }
            }
        }
    }

    static int fget_input_char(InputStream in, ParseOptions opts) throws IOException {
        if (!opts.parse_options_get_batch_mode() && (opts.verbosity > 0))
            opts.out.print("linkparser> ");
        opts.out.flush();
        return in.read();
    }

    static boolean fget_input_string(StringBuffer input_string, InputStream in, PrintStream out, ParseOptions opts)
        throws IOException {
        int c;
        input_string.setLength(0);
        if (input_pending) {
            input_pending = false;
            c = input_char;
        } else {
            if (!opts.parse_options_get_batch_mode() && opts.verbosity > 0)
                out.println("linkparser> ");
            out.flush();
            c = in.read();
        }
        while (c != '\n') {
            if (c < 0) {
                return false;
            }
            input_string.append((char)c);
            c = in.read();
        }
        return true;
    }

}
