/* ====================================================================
 * The Apache Software License, Version 1.1
 *
 * Copyright (c) 2003 The Apache Software Foundation.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution, if
 *    any, must include the following acknowlegement:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowlegement may appear in the software itself,
 *    if and wherever such third-party acknowlegements normally appear.
 *
 * 4. The names "The Jakarta Project", "Commons", and "Apache Software
 *    Foundation" must not be used to endorse or promote products derived
 *    from this software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache"
 *    nor may "Apache" appear in their names without prior written
 *    permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 */
package org.apache.commons.text.spell;

import java.util.Iterator;
import java.util.List;
import java.util.Map;

/**
 * This implementation composes words from dictionary entries
 * and selected prefixes and suffixes.
 * 
 *
 */
public class AffixWordComposer implements WordComposer {
    private Dictionary dictionary;
    private Map prefix;
    private Map suffix;

    public AffixWordComposer(Map prefix, Map suffix, Dictionary dictionary) {
        this.dictionary = dictionary;
        this.prefix = prefix;
        this.suffix = suffix;
    }
    /* (non-Javadoc)
     * @see org.apache.commons.text.spell.WordComposer#lookup(java.lang.String)
     */
    public DictionaryEntry lookup(String word) {
        DictionaryEntry entry = dictionary.lookup(word);
        if (entry != null) {
            return entry;
        }
        Map m = prefix;
        int i = 0;
        while (m != null && i < word.length()) {
            List prefixes = (List)m.get("");
            if (prefixes != null) {
                String strippedWord = word.substring(i);
                System.out.println(
                    " got s=" + prefixes.size() + " p=" + word.substring(0, i) + " word=" + strippedWord);
                for (Iterator it = prefixes.iterator(); it.hasNext();) {
                    PrefixEntry prefix = (PrefixEntry)it.next();
                    String checkWord = strippedWord;
                    if (prefix.strip != null) {
                        checkWord = prefix.strip + strippedWord;
                    }
                    System.out.println(" check word=" + checkWord);
                    if (prefix.checkConditions(checkWord)) {
                        System.out.println(" check ok");
                        entry = dictionary.lookup(checkWord);
                        if (entry != null) {
                            if (entry.getAffix().indexOf(prefix.affix) >= 0) {
                                System.out.println(" found!");
                                return entry;
                            } else {
                                System.out.println(" aff mismatch! a=" + entry.getAffix() + " x=" + prefix.affix);
                            }
                        }
                        if (prefix.compoundFlag) {
                            entry = checkSuffix(checkWord, true);
                            if (entry != null) {
                                System.out.println(" found compound!");
                                return entry;
                            }
                        }
                    }
                }
            }
            m = (Map)m.get(word.substring(i, i + 1));
            i++;
        }
        return checkSuffix(word, false);
    }

    private DictionaryEntry checkSuffix(String word, boolean compound) {
        Map m = suffix;
        int i = word.length();
        while (m != null && i > 0) {
            List suffixes = (List)m.get("");
            if (suffixes != null) {
                String strippedWord = word.substring(0, i);
                System.out.println(" got s=" + suffixes.size() + " p=" + word.substring(i) + " word=" + strippedWord);
                for (Iterator it = suffixes.iterator(); it.hasNext();) {
                    SuffixEntry suffix = (SuffixEntry)it.next();
                    System.out.println(" a=" + suffix.affix + " strip=" + suffix.strip + " numc=" + suffix.numconds);
                    if (!compound || suffix.compoundFlag) {
                        String checkWord = strippedWord;
                        if (suffix.strip != null) {
                            checkWord = strippedWord + suffix.strip;
                        }
                        System.out.println(" check word=" + checkWord);
                        if (suffix.checkConditions(checkWord)) {
                            System.out.println(" check ok");
                            DictionaryEntry entry = dictionary.lookup(checkWord);
                            if (entry != null) {
                                if (entry.getAffix().indexOf(suffix.affix) >= 0) {
                                    System.out.println(" found!");
                                    return entry;
                                } else {
                                    System.out.println(" aff mismatch! a=" + entry.getAffix() + " x=" + suffix.affix);
                                }
                            }
                        }
                    }
                }
            }
            m = (Map)m.get(word.substring(i - 1, i));
            i--;
        }
        return null;
    }

}
