package org.apache.commons.text.util;

import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.Reader;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.StringTokenizer;
import java.util.TreeMap;
import java.util.TreeSet;
import java.util.Set;

/**
 * TODO add javadoc
 */
public class MorphGen {

    /**
     * TODO add javadoc 
     */
    public MorphGen() {
        super();
    }

    private static final int BUFSIZE = 1024;
    private static final int WORDSIZE = 80;

    private static final int STATE_OTHER = 0;
    private static final int STATE_WORD = 1;
    private static final int STATE_MARKUP = 2;

    public static void main(String arg[]) {
        //tokenizeCorups();
        readMorph();
        if (root.tail != null) {
            System.out.println(root.tail.size());
        } else {
            System.out.println("null");
        }
        lookup("eines");
        lookup("eins");
    }

    private static class TrieEntry {
        Map tail;
        List values;
    }

    private static class MorphEntry {
        String morphClass;
        List derivations;

        MorphEntry(String morphClass, List derivations) {
            this.morphClass = morphClass;
            this.derivations = derivations;
        }
    }

    static TrieEntry root = null;

    public static void readLex() {
        try {
            BufferedReader reader = new BufferedReader(new FileReader("data/lex/lex.data"));
            int lineNr = 1;
            String line = reader.readLine();
            while (line != null) {
                line = line.trim();
                if (line.length() > 0) {

                }
                lineNr++;
                line = reader.readLine();
            }
            reader.close();
        } catch (FileNotFoundException e) {
            System.out.println("File not found " + e);
        } catch (IOException e) {
            System.out.println("IO exception " + e);
        }
    }

    public static void readMorph() {
        try {
            root = new TrieEntry();
            BufferedReader reader = new BufferedReader(new FileReader("data/lex/morph.data"));
            int lineNr = 1;
            String line = reader.readLine();
            int columnNr = -1;
            List columnDescriptions = null;
            while (line != null) {
                line = line.trim();
                if (line.length() > 0) {
                    StringTokenizer tok = new StringTokenizer(line);
                    String token = tok.nextToken();
                    if (token.equals("table")) {
                        columnNr = 0;
                        columnDescriptions = new ArrayList();
                    } else if (Character.isDigit(token.charAt(0))) {
                        // column definition or column number row
                        try {
                            int i = Integer.parseInt(token);
                            if (i != columnNr) {
                                // column number row or unexpected column definition row, column number not increasing
                                if (i != 0) {
                                    // ouch, not a column number row
                                    throw new RuntimeException(lineNr + ": Unexpected token " + token);
                                }
                                // column number row, start data table
                                // sanity check column headers
                                int j = 1;
                                while (tok.hasMoreTokens() && j < columnNr) {
                                    int k = Integer.parseInt(tok.nextToken());
                                    if (k != j) {
                                        throw new RuntimeException(
                                            lineNr + ": Unexpected column header " + k + " in column " + j);
                                    }
                                    j++;
                                }
                                if (tok.hasMoreTokens()) {
                                    throw new RuntimeException(
                                        lineNr + ": Trailing columns: " + tok.nextToken() + " ...");
                                }
                                if (j < columnNr) {
                                    throw new RuntimeException(lineNr + ": Missing columns, " + j + " columns found");
                                }
                            } else {
                                // column definition row
                                List columnDescription = new ArrayList();
                                while (tok.hasMoreTokens()) {
                                    token = tok.nextToken();
                                    columnDescription.add(token);
                                }
                                columnDescriptions.add(columnDescription);
                                columnNr++;
                            }
                        } catch (NumberFormatException e1) {
                            throw new RuntimeException(lineNr + ": Unexpected token " + token);
                        }
                    } else {
                        // data row
                        String morphClass = token;
                        int j = 0;
                        while (tok.hasMoreTokens() && j < columnNr) {
                            token = tok.nextToken();
                            // insert suffix
                            TrieEntry entry = root;
                            if (!token.equals("0")) {
                                boolean optional = false;
                                int k = token.length() - 1;
                                while (k >= 0) {
                                    if (token.charAt(k) == ']') {
                                        optional = true;
                                        if (entry.values == null) {
                                            entry.values = new ArrayList();
                                        }
                                        entry.values.add(new MorphEntry(morphClass, (List)columnDescriptions.get(j)));
                                        k--;
                                        continue;
                                    } else if (token.charAt(k) == '[') {
                                        if (!optional) {
                                            throw new RuntimeException(
                                                lineNr + ": unexpected '[': " + token + " in column " + j);
                                        }
                                        if (k != 0) {
                                            throw new RuntimeException(
                                                lineNr
                                                    + ": '[' must be at start of suffix: "
                                                    + token
                                                    + " in column "
                                                    + j);
                                        }
                                        break;
                                    } else if (!Character.isLetter(token.charAt(k))) {
                                        if (k != 0) {
                                            throw new RuntimeException(
                                                lineNr
                                                    + ": stem handling flags must be at start of suffix: "
                                                    + token
                                                    + " in column "
                                                    + j);
                                        }
                                    }
                                    if (entry.tail == null) {
                                        entry.tail = new TreeMap();
                                    }
                                    TrieEntry child = (TrieEntry)entry.tail.get(new Character(token.charAt(k)));
                                    if (child == null) {
                                        child = new TrieEntry();
                                        entry.tail.put(new Character(token.charAt(k)), child);
                                    }
                                    k--;
                                    entry = child;
                                }
                            }
                            if (entry.values == null) {
                                entry.values = new ArrayList();
                            }
                            entry.values.add(new MorphEntry(morphClass, (List)columnDescriptions.get(j)));
                            j++;
                        }
                        if (tok.hasMoreTokens()) {
                            throw new RuntimeException(lineNr + ": Trailing columns: " + tok.nextToken() + " ...");
                        }
                        if (j < columnNr) {
                            throw new RuntimeException(lineNr + ": Missing columns, " + (j + 1) + " columns found");
                        }
                    }
                } else {
                    // end table
                    columnNr = -1;
                }
                lineNr++;
                line = reader.readLine();
            }
            reader.close();
        } catch (FileNotFoundException e) {
            System.out.println("File not found " + e);
        } catch (IOException e) {
            System.out.println("IO exception " + e);
        }
    }

    public static void outputDerivations(String word, int stemEnd, List morphs) {
        String stem = word.substring(0, stemEnd + 1);
        for (Iterator iter = morphs.iterator(); iter.hasNext();) {
            MorphEntry element = (MorphEntry)iter.next();
            System.out.print(stem + ": " + element.morphClass);
            for (Iterator iterator = element.derivations.iterator(); iterator.hasNext();) {
                System.out.print(' ' + (String)iterator.next());
            }
            System.out.println();
        }
    }

    public static void lookup(String word) {
        TrieEntry entry = root;
        int k = word.length() - 1;
        while (k >= 0) {
            if (entry.values != null) {
                outputDerivations(word, k, entry.values);
            }
            if (entry.tail == null) {
                System.out.println("---");
                return;
            }
            TrieEntry child = (TrieEntry)entry.tail.get(new Character(word.charAt(k)));
            if (child == null) {
                System.out.println("---");
                return;
            }
            k--;
            entry = child;
        }
        if (entry.values != null) {
            outputDerivations(word, k, entry.values);
        }
        System.out.println("---");
    }

    public static void tokenizeCorpus() {
        try {
            Reader reader = new FileReader("data/corpus/ep-00-09-21.txt");
            PrintWriter writer = new PrintWriter(new FileWriter("data/corpus/words000.txt"));
            char buffer[] = new char[BUFSIZE];
            StringBuffer word = new StringBuffer(32);
            Set words = new TreeSet();
            int i;
            int state = STATE_OTHER;
            while ((i = reader.read(buffer)) >= 0) {
                for (int j = 0; j < i; j++) {
                    char c = buffer[j];
                    switch (c) {
                        case '<' :
                            if (state == STATE_WORD) {
                                words.add(word.toString());
                                word.setLength(0);
                            }
                            state = STATE_MARKUP;
                            break;
                        case '>' :
                            if (state != STATE_MARKUP) {
                                System.out.println("unexpected '>'");
                            }
                            state = STATE_OTHER;
                            break;
                        case 'A' :
                        case 'B' :
                        case 'C' :
                        case 'D' :
                        case 'E' :
                        case 'F' :
                        case 'G' :
                        case 'H' :
                        case 'I' :
                        case 'J' :
                        case 'K' :
                        case 'L' :
                        case 'M' :
                        case 'N' :
                        case 'O' :
                        case 'P' :
                        case 'Q' :
                        case 'R' :
                        case 'S' :
                        case 'T' :
                        case 'U' :
                        case 'V' :
                        case 'W' :
                        case 'X' :
                        case 'Y' :
                        case 'Z' :
                        case '' :
                        case '' :
                        case '' :
                        case '' :
                        case '' :
                        case '' :
                        case '' :
                        case '' :
                        case '' :
                        case '' :
                        case '' :
                        case '' :
                        case '' :
                        case '' :
                        case '' :
                        case '' :
                        case '' :
                        case '' :
                        case '' :
                        case '' :
                        case '' :
                        case '' :
                        case '' :
                        case '' :
                        case '' :
                            c = Character.toLowerCase(c);
                        case 'a' :
                        case 'b' :
                        case 'c' :
                        case 'd' :
                        case 'e' :
                        case 'f' :
                        case 'g' :
                        case 'h' :
                        case 'i' :
                        case 'j' :
                        case 'k' :
                        case 'l' :
                        case 'm' :
                        case 'n' :
                        case 'o' :
                        case 'p' :
                        case 'q' :
                        case 'r' :
                        case 's' :
                        case 't' :
                        case 'u' :
                        case 'v' :
                        case 'w' :
                        case 'x' :
                        case 'y' :
                        case 'z' :
                        case '' :
                        case '' :
                        case '' :
                        case '' :
                        case '' :
                        case '' :
                        case '' :
                        case '' :
                        case '' :
                        case '' :
                        case '' :
                        case '' :
                        case '' :
                        case '' :
                        case '' :
                        case '' :
                        case '' :
                        case '' :
                        case '' :
                        case '' :
                        case '' :
                        case '' :
                        case '' :
                        case '' :
                        case '' :
                        case '' :
                            if (state == STATE_OTHER || state == STATE_WORD) {
                                state = STATE_WORD;
                                word.append(c);
                            }
                            break;
                        default :
                            System.out.println("unexpected " + c + " (#" + Integer.toHexString(c) + ")");
                        case ' ' :
                        case '\t' :
                        case '\n' :
                        case '\r' :
                        case ',' :
                        case '.' :
                        case '(' :
                        case ')' :
                        case '?' :
                        case '!' :
                        case '=' :
                        case '"' :
                        case '0' :
                        case '9' :
                        case '8' :
                        case '7' :
                        case '6' :
                        case '5' :
                        case '4' :
                        case '3' :
                        case '2' :
                        case '1' :
                        case '-' :
                        case ':' :
                        case '/' :
                        case '%' :
                        case ';' :
                        case '' :
                            if (state == STATE_WORD) {
                                words.add(word.toString());
                                word.setLength(0);
                                state = STATE_OTHER;
                            }
                            break;
                    }
                }
            }
            reader.close();
            if (word.length() > 0) {
                words.add(word.toString());
            }
            for (Iterator iter = words.iterator(); iter.hasNext();) {
                String element = (String)iter.next();
                writer.println(element);
                System.out.println(element);
            }
            writer.close();
        } catch (FileNotFoundException e) {
            System.out.println("File not found " + e);
        } catch (IOException e) {
            System.out.println("IO exception " + e);
        }
    }
}
